import numpy as np
from numpy.fft import fftn, ifftn
import warnings
from .expected_periodogram import autocov


def sim_circ_embedding(cov_func, shape):
    cov = autocov(cov_func, shape)
    m, n = shape
    f = np.real(4 * m * n * fftn(cov))
    min_ = np.min(f)
    if min_ <= 0:
        warnings.warn(f'Embedding is not positive definite, min value {min_}.')
    e = (np.random.randn(*f.shape) + 1j * np.random.randn(*f.shape))
    z = np.sqrt(np.maximum(f, 0)) * e
    z_inv = ifftn(z)
    return np.real(z_inv[:shape[0], :shape[1]]), min_



import numpy as np
from numpy.fft import fftn, ifftn


def spatial_kernel(data):
    """Compute the spatial kernel, cg in the paper, via FFT.

    :param data: np.ndarray, the data for which we want to compute the sample autocovariance
    In the case of a full grid with g=1 everywhere on the grid this takes a simple analytical form."""
    n = list(data.shape)
    two_n = list(data.shape)
    for i in range(len(two_n)):
        two_n[i] = two_n[i] * 2 - 1
    f = abs(fftn(data, two_n))**2
    slices = tuple(slice(None, 2 * s - 1, None) for s in n)
    cg = np.real(ifftn(f)[slices])
    # Normalise as in the paper
    cg /= cg.flatten()[0]
    return cg



__version__ = '0.1.1'

from .likelihood import fit, periodogram
from .cov_funcs import exp_cov, sq_exp_cov, matern15_cov_func, exp_cov2, matern
from .expected_periodogram import compute_ep
from .simulation import sim_circ_embedding


import warnings
import numpy as np
from numpy.fft import fftn
from scipy.optimize import fmin_l_bfgs_b
from scipy.signal import hanning
from .expected_periodogram import compute_ep


def prod_list(l):
    if len(l) == 0:
        return 1
    else:
        return l[0] * prod_list(l[1:])


def shape_two(shape):
    new_shape = []
    for e in shape:
        new_shape.append(2 * e)
    return tuple(new_shape)


def periodogram(y, grid=None, fold=True):
    shape = y.shape
    if grid is None:
        n = prod_list(shape)
    else:
        n = np.sum(grid**2)
    if not fold:
        shape = shape_two(shape)
    return 1 / n * abs(fftn(y, shape))**2


def apply_taper(y):
    n_1, n_2 = y.shape
    taper_1 = hanning(n_1).reshape((n_1, 1))
    taper_2 = hanning(n_2).reshape((1, n_2))
    taper = np.dot(taper_1, taper_2)
    return y * taper, taper


def whittle(per, e_per):
    n = prod_list(per.shape)
    return 1 / n * np.sum(np.log(e_per) + per / e_per)


def whittle_prime(per, e_per, e_per_prime):
    n = prod_list(per.shape)
    return 1 / n * np.sum((e_per - per) * e_per_prime / e_per ** 2)


def fit(y, grid, cov_func, init_guess, fold=True, cov_func_prime=None, taper=False):
    # apply taper if required
    if taper:
        y, taper = apply_taper(y)
        grid = grid * taper

    per = periodogram(y, grid, fold=fold)

    if cov_func_prime is not None:
        def opt_func(x):
            e_per = compute_ep(lambda lags: cov_func(lags, *x), grid, fold=fold)
            e_per_prime = compute_ep(lambda lags: (cov_func_prime(lags, *x))[0], grid, fold=fold)
            return whittle(per, e_per), whittle_prime(per, e_per, e_per_prime)

        est_, fval, info = fmin_l_bfgs_b(lambda x: opt_func(x),
                                         init_guess,
                                         maxiter=100,
                                         bounds=[(0.01, 1000), ] * len(init_guess))
    else:
        def opt_func(x):
            e_per = compute_ep(lambda lags: cov_func(lags, *x), grid, fold=fold)
            return whittle(per, e_per)
        est_, fval, info = fmin_l_bfgs_b(lambda x: opt_func(x),
                                         init_guess,
                                         maxiter=100,
                                         approx_grad=True,
                                         bounds=[(0.01, 1000), ] * len(init_guess))

    if info['warnflag'] != 0:
        warnings.warn('Issue during optimization')

    return est_


import numpy as np
from numpy.fft import fft, ifft, fftn, ifftn, fftshift, ifftshift
from .spatial_kernel import spatial_kernel


def autocov(cov_func, shape):
    """Compute the covariance function on a grid of lags determined by the passed shape.
    In d=1 the lags would be -(n-1)...n-1, but then a iffshit is applied so that the lags are
    0 ... n-1 -n+1 ... -1. This may look weird but it makes it easier to do the folding operation
    when computing the expecting periodogram"""
    xs = np.meshgrid(*(np.arange(-n+1, n) for n in shape), indexing='ij')
    return ifftshift(cov_func(xs))


def compute_ep(cov_func, grid, fold=True):
    """Computes the expected periodogram, for the passed covariance function and grid. The grid is an array, in
    the simplest case just an array of ones
    :param cov_func: covariance function
    :param grid: array
    :param fold: True if we compute the expected periodogram on the natural Fourier grid, False if we compute it on a
    frequency grid with twice higher resolution
    :return: """
    shape = grid.shape
    n_dim = grid.ndim
    # In the case of a complete grid, cg takes a closed form.
    cg = spatial_kernel(grid)
    acv = autocov(cov_func, shape)
    cbar = cg * acv
    # now we need to "fold"
    if fold:
        result = np.zeros_like(grid)
        if n_dim == 1:
            for i in range(2):
                result[i:] += cbar[i * shape[0]: (i + 1) * shape[0]]
        elif n_dim == 2:
            for i in range(2):
                for j in range(2):
                    result[i:, j:] += cbar[i * shape[0]: (i + 1) * shape[0], j * shape[1]: (j + 1) * shape[1]]
        elif n_dim == 3:
            for i in range(2):
                for j in range(2):
                    for k in range(2):
                        result[i:, j:, k:] += cbar[i * shape[0]: (i + 1) * shape[0],
                                              j * shape[1]: (j + 1) * shape[1],
                                              k * shape[2]: (k + 1) * shape[2]]
    else:
        m, n = shape
        result = np.zeros((2 * m, 2 * n))
        result[:m, :n] = cbar[:m, :n]
        result[m+1:, :n] = cbar[m:, :n]
        result[m+1:, n+1:] = cbar[m:, n:]
        result[:m, n+1:] = cbar[:m, n:]
    # We take the real part of the fft only due to numerical precision, in theory this should be real
    result = np.real(fftn(result))
    return result



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sdwl']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.1.2,<4.0.0', 'numpy>=1.21.5,<2.0.0', 'scipy>=1.7.3,<2.0.0']

setup_kwargs = {
    'name': 'sdwl',
    'version': '0.1.1',
    'description': '',
    'long_description': None,
    'author': 'arthur',
    'author_email': 'ahw795@qmul.ac.uk',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'http://arthurpgb.pythonanywhere.com/sdw',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



from abc import ABC, abstractmethod
from pathlib import Path
import matplotlib.pyplot as plt

import numpy as np

from typing import Tuple

PATH_TO_FRANCE_IMG = str(Path(__file__).parents[1] / 'france.jpg')

class Grid(ABC):
    def __init__(self, shape: Tuple[int]):
        self.shape = shape

    @abstractmethod
    def get_new(self):
        pass

    def __mul__(self, other):
        return GridProduct(self, other)


class GridProduct(Grid):
    def __init__(self, grid1, grid2):
        assert grid1.shape == grid2.shape
        shape = grid1.shape
        super(GridProduct, self).__init__(shape)
        self.grid1 = grid1
        self.grid2 = grid2

    def get_new(self):
        return self.grid1.get_new() * self.grid2.get_new()


class FullGrid(Grid):
    def get_new(self):
        return np.ones(self.shape)


class CircleGrid(Grid):
    def __init__(self, shape: Tuple[int], center: Tuple[int], diameter: float):
        super().__init__(shape)
        self.center = center
        self.diameter = diameter

    def get_new(self):
        (x_0, y_0), diameter = self.center, self.diameter
        shape = self.shape
        x, y = np.meshgrid(np.arange(shape[0]), np.arange(shape[1]), indexing='ij')
        circle = ((x - x_0) ** 2 + (y - y_0) ** 2) <= 1 / 4 * diameter ** 2
        circle = circle * 1.
        return circle


class BernoulliGrid(Grid):
    def __init__(self, shape: Tuple[int], p: float):
        super(BernoulliGrid, self).__init__(shape)
        self.p = p

    @property
    def p(self):
        return self._p

    @p.setter
    def p(self, value: float):
        assert 0 <= value <= 1, 'p should be a probability'
        self._p = value

    def get_new(self):
        epsilon = np.random.rand(*self.shape)
        return (epsilon >= self.p) * 1.


class ImgGrid(Grid):
    def __init__(self, shape, img_path: str = PATH_TO_FRANCE_IMG):
        super().__init__(shape)
        self.img_path = img_path
        img = plt.imread(self.img_path)
        img = (img[110:-110, 110:-110, 0] == 0) * 1.
        self.img = np.flipud(img)

    @property
    def img(self):
        return self._img

    @img.setter
    def img(self, value):
        self._img = value

    def interpolate(self):
        m_0, n_0 = self.img.shape
        m, n = self.shape
        x = np.asarray(np.arange(n) / n * n_0, np.int64)
        y = np.asarray(np.arange(m) / m * m_0, np.int64)
        xx, yy = np.meshgrid(x, y, indexing='xy')
        return self.img[yy, xx]

    def get_new(self):
        return self.interpolate()



import numpy as np
from scipy.special import gamma, kv

def exp_cov(lags, rho, sigma=1.):
     return sigma ** 2 * np.exp(-np.sqrt(sum((lag**2 for lag in lags))) / rho)


def exp_cov2(lags, rho_1, rho_2, theta=0., sigma=1.):
    lag_0 = lags[0] * np.cos(theta) - lags[1] * np.sin(theta)
    lag_1 = lags[0] * np.sin(theta) + lags[1] * np.cos(theta)
    norm = np.sqrt((lag_0 / rho_1)**2 + (lag_1 / rho_2)**2)
    return sigma**2 * np.exp(-norm)


def matern15_cov_func(lags, rho, sigma=1.):
    K = np.sqrt(3) * np.sqrt(sum((lag**2 for lag in lags))) / rho
    return (1.0 + K) * np.exp(-K)


def sq_exp_cov(lags, rho, sigma=1.):
    lags = np.stack(lags)
    return sigma ** 2 * np.exp(-(sum((lag**2 for lag in lags))) / (2 * rho**2)) + 0.001 * (np.all(lags == 0, axis=0))


def matern(lags, rho, nu, sigma=1.):
    d = np.sqrt(sum((lag**2 for lag in lags)))
    term1 = 2 ** (1 - nu) / gamma(nu)
    term2 = (np.sqrt(2 * nu) * d / rho) ** nu
    term3 = kv(nu, np.sqrt(2 * nu) * d / rho)
    val = sigma ** 2 * term1 * term2 * term3
    val[d == 0] = sigma ** 2
    return val


## Derivatives

def exp_cov_prime(lags, rho, sigma=1.):
    d_rho = exp_cov(lags, rho, sigma) * np.sqrt(sum((lag**2 for lag in lags))) / rho ** 2
    return (d_rho, )


def sq_exp_cov_prime(lags, rho, sigma=1.):
    d_rho = sq_exp_cov(lags, rho, sigma) * (sum((lag**2 for lag in lags))) / rho ** 3
    return (d_rho, )

def matern15_cov_func_prime(lags, rho, sigma=1.):
    d_rho = matern15_cov_func(lags, rho, sigma) * np.sqrt(3) * np.sqrt(sum((lag**2 for lag in lags))) / rho ** 2
    return (d_rho,)